﻿using Microsoft.Win32;
using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Linq;
using System.Text;
using System.Threading;
using System.Threading.Tasks;
using System.Windows.Forms;

namespace nRF24L01P
{
    public partial class Form1 : Form
    {
        public string port_name = null;
        public Form1()
        {
            this.StartPosition = FormStartPosition.CenterScreen;
            InitializeComponent();
        }

        //接收数据缓冲区
        public Byte[] RecvBuf = new Byte[4096];
        //接收数据计数器
        public int RecvCnt = 0;

        private void ComBox_MouseClick(object sender, MouseEventArgs e)
        {
            int num = 0;
            ComBox.Items.Clear();
            RegistryKey keyCom = Registry.LocalMachine.OpenSubKey("Hardware\\DeviceMap\\SerialComm");
            if (keyCom != null)
            {
                string[] sSubKeys = keyCom.GetValueNames();
                this.ComBox.Items.Clear();
                foreach (string sName in sSubKeys)
                {
                    num++;
                    string sValue = (string)keyCom.GetValue(sName);
                    this.ComBox.Items.Add(sValue);
                }

            }
            else
            {
                MessageBox.Show("未发现串口");
            }
        }

        //RCC校验
        public Byte RCC(byte[] buf,int index,int length)
        {
            Byte code = 0x00;
            int cnt = 0;
            while (cnt < length)
            {
                code ^= buf[index + cnt];
                cnt++;
            }
            return code;
        }

        //查找有效报文位置
        public int CheckRxText()
        {
            int cntsta;
            if (RecvCnt < 5)
            {
                //报文至少7个数据
                return -3;  //报文不全
            }
            for (cntsta = 0; cntsta < RecvCnt; cntsta++)
            {
                if (RecvBuf[cntsta] == 0xaa)
                {
                    if (RecvBuf[cntsta + 1] == 0x55)
                    {
                        //AA55起始码
                        if ((cntsta + 5 + RecvBuf[cntsta + 2]) <= RecvCnt)
                        {
                            if (RecvBuf[(cntsta + 3 + RecvBuf[cntsta + 2])] == 0x55)
                            {
                                if (RecvBuf[(cntsta + 4 + RecvBuf[cntsta + 2])] == 0xaa)
                                {
                                    //55AA结束码，校验RCC
                                    if (RecvBuf[cntsta + RecvBuf[cntsta + 2] + 2] == RCC(RecvBuf, cntsta + 2, RecvBuf[cntsta + 2]))
                                    {
                                        return cntsta;
                                    }
                                }
                            }
                        }
                        else
                        {
                            //报文不全
                            return -3;
                        }
                    }
                }
            }
            return -1;  //报文错误
        }

        //声明一条进程，用来接收串口数据
        Thread RecvThread;
        //声明一个回调，用于主进程更新接收到的数据
        delegate void RecvInvoke();
        //声明一个回调，用于连接异常时断开连接
        delegate void RecvError();

        //创建回调函数
        public void RecvData()
        {
            if (DataCntNRF.Checked == false)
            {
                RxCntLable.Text = (Convert.ToInt32(RxCntLable.Text) + 1).ToString();
            }
            int cnt;
            cnt = CheckRxText();
            if (cnt < 0)
            {
                //报文有误
                return;
            }
            //有正常报文
            Byte[] dat = new Byte[RecvBuf[cnt + 2]];
            //截取有效报文
            for (int i = 0; i < dat.Length; i++)
            {
                dat[i] = RecvBuf[cnt + 2 + i];
            }
            RecvCnt = 0;    //清零RX缓冲
            switch (dat[1])
            {
                case 0x00:
                    //重发上一帧数据
                    break;
                case 0x01:
                    //发送数据,上位机专属命令
                    break;
                case 0x02:
                    //接收到数据
                    if (DataCntNRF.Checked)
                    {
                        //nrf接收到数据包
                        RxCntLable.Text = (Convert.ToInt32(RxCntLable.Text) + 1).ToString();
                    }
                    if (dat[0] <= 2)
                    {
                        //数据错误，进行错误处理
                    }
                    else
                    {
                        //收到数据，一般32字节
                        cnt = Convert.ToInt32(dat[0]) - 2;
                        if (HexModeCheck.Checked)
                        {
                            //以16进制显示
                            RxText.Text += DateTime.Now.ToLongTimeString().ToString();
                            RxText.Text += ":\r\n";
                            for (int i = 0; i < cnt; i++)
                            {
                                RxText.Text += dat[i + 2].ToString("X2").ToUpper();
                                RxText.Text += " ";
                            }
                        }
                        else
                        {
                            //以文本显示
                            RxText.Text += DateTime.Now.ToLongTimeString().ToString();
                            RxText.Text += ":\r\n";
                            try
                            {
                                for (int i = 2; i < dat.Length; i++)
                                {
                                    dat[i - 2] = dat[i];
                                }
                                dat[dat.Length - 2] = 0x00;
                                dat[dat.Length - 1] = 0x00;
                                RxText.Text += Encoding.GetEncoding(TextEncodeBox.Text).GetString(dat);
                            } catch
                            {
                                TxText.Text += "无法解码字符串:\r\n";
                                for (int i = 0; i < dat.Length - 2; i++)
                                {
                                    RxText.Text += dat[i + 2].ToString("X2").ToUpper();
                                }
                            }
                        }
                        RxText.Text += "\r\n";
                    }
                    break;
                case 0x03:
                    //修改配置
                    RxText.Text += "修改配置中\r\n";
                    break;
                case 0x04:
                    //读取配置
                    if (dat[0] == 26)
                    {
                        //有效数据应是26个
                        if (dat[2] != 0xff)
                        {
                            //发射通道
                            if (ModeRFchHex.Checked)
                                ModeRFchBox.Text = dat[2].ToString("X2").ToUpper();
                            else
                                ModeRFchBox.Text = dat[2].ToString();
                        }
                        //校验方式
                        if (dat[3] == 0x0b)
                        {
                            ModeCRCxBox.SelectedIndex = 0;
                        }
                        else if (dat[3] == 0x0f)
                        {
                            ModeCRCxBox.SelectedIndex = 1;
                        }
                        else
                        {
                            ModeCRCxBox.Text = "";
                        }

                        //nRF_RF_SETUP
                        if (dat[4] == 0x07)
                        {
                            ModeRFSpeedBox.SelectedIndex = 0;
                        }
                        else if (dat[4] == 0x0f)
                        {
                            ModeRFSpeedBox.SelectedIndex = 1;
                        }
                        else
                        {
                            ModeRFSpeedBox.Text = "";
                        }
                        
                        //发送地址
                        ModeTxAddrBox.Text = "";
                        for (int i = 6; i < 11; i++)
                        {
                            ModeTxAddrBox.Text += dat[i].ToString("X2").ToUpper();
                        }
                        //接收地址
                        ModeRxAddrBox.Text = "";
                        for (int i = 12; i < 17; i++)
                        {
                            ModeRxAddrBox.Text += dat[i].ToString("X2").ToUpper();
                        }
                        //波特率
                        if (dat[17] > 8)
                        {
                            ModeBaudBox.Text = "";
                        }
                        ModeBaudBox.SelectedIndex = dat[17];
                        //nRF_RETR
                        SetupRetrH.SelectedIndex = Convert.ToInt32(dat[18] / 0xf);
                        SetupRetrL.SelectedIndex = Convert.ToInt32(dat[18] % 0xf);
                        //序列号
                        ModeIDBox.Text = "";
                        for (int i = 19; i < 26; i++)
                        {
                            ModeIDBox.Text += dat[i].ToString("X2").ToUpper();
                        }
                    }
                    break;
                case 0x05:
                    //发送成功
                    TxText.Text += "发送成功\r\n";
                    if (DataCntNRF.Checked)
                    {
                        TxCntLable.Text = (Convert.ToInt32(TxCntLable.Text) + 1).ToString();
                    }
                    break;
                case 0x06:
                    //发送超时
                    TxText.Text += "发送超时\r\n";
                    break;
                case 0x07:
                    //发送失败
                    TxText.Text += "发送失败\r\n";
                    break;
                case 0x08:
                    RxText.Text += "设备正在重启\r\n";
                    //重启设备
                    break;
                case 0x09:
                    //重置设备
                    RxText.Text += "设备重置中\r\n";
                    break;
                case 0x0a:
                    RxText.Text += "命令有误\r\n";
                    break;
                default:
                    RxText.Text += "命令有误\r\n";
                    break;
            }

            RecvCnt = 0;
        }

        //错误回调函数
        public void RxError()
        {
            try
            {
                serialPort1.Close();
            }
            catch
            { }
            UartConn.Text = "连接";
            ComBox.Enabled = true;
            ComBaudBox.Enabled = true;
            ModeWriteConf.Enabled = false;
            ModeReadConf.Enabled = false;
            ModeReset.Enabled = false;
            ModeRst.Enabled = false;
            SendTxBuf.Enabled = false;
        }



        //使用串口发送数据
        public void UartSend(byte[] buf)
        {
            byte[] dat = new Byte[buf.Length + 5];
            dat[0] = 0xaa;  //启动码
            dat[1] = 0x55;
            dat[buf.Length + 3] = 0x55; //结束码
            dat[buf.Length + 4] = 0xaa;
            dat[buf.Length + 2] = RCC(buf,0,buf.Length); //校验码
            for (int cnt = 0; cnt < buf.Length; cnt++)
            {
                dat[cnt + 2] = buf[cnt];
            }
            serialPort1.Write(dat, 0, dat.Length);
            if (DataCntNRF.Checked == false)
            {
                TxCntLable.Text = (Convert.ToInt32(TxCntLable.Text) + 1).ToString();
            }
        }

        private void Form1_Load(object sender, EventArgs e)
        {
            ComBaudBox.Items.Add("1200");
            ComBaudBox.Items.Add("2400");
            ComBaudBox.Items.Add("4800");
            ComBaudBox.Items.Add("9600");
            ComBaudBox.Items.Add("19200");
            ComBaudBox.Items.Add("38400");
            ComBaudBox.Items.Add("57600");
            ComBaudBox.Items.Add("115200");
            ComBaudBox.SelectedIndex = ComBaudBox.Items.IndexOf("115200");
            ModeBaudBox.Items.Add("1200");
            ModeBaudBox.Items.Add("2400");
            ModeBaudBox.Items.Add("4800");
            ModeBaudBox.Items.Add("9600");
            ModeBaudBox.Items.Add("19200");
            ModeBaudBox.Items.Add("38400");
            ModeBaudBox.Items.Add("57600");
            ModeBaudBox.Items.Add("115200");
            ModeRFSpeedBox.Items.Add("1Mbps");
            ModeRFSpeedBox.Items.Add("2Mbps");
            ModeCRCxBox.Items.Add("CRC8");
            ModeCRCxBox.Items.Add("CRC16");
            TextEncodeBox.Items.Add("gb2312");
            TextEncodeBox.Items.Add("unicode");
            TextEncodeBox.Items.Add("gbk");
            TextEncodeBox.Items.Add("utf-8");
            TextEncodeBox.SelectedIndex = TextEncodeBox.Items.IndexOf("gb2312");
            TextEncodeBox.Enabled = false;
            for (int i = 0; i < 16; i++)
            {
                SetupRetrL.Items.Add(i.ToString());
                SetupRetrH.Items.Add("250*"+i.ToString()+"+86");
            }
            SetupRetrL.SelectedIndex = 10;
            SetupRetrH.SelectedIndex = 1;
        }

        private void CleanRxText_Click(object sender, EventArgs e)
        {
            RxText.Text = "";
        }

        private void CleanTxText_Click(object sender, EventArgs e)
        {
            TxText.Text = "";
        }

        private void CleanTxBuf_Click(object sender, EventArgs e)
        {
            TxBufText.Text = "";
        }

        private void CleanCnt_Click(object sender, EventArgs e)
        {
            RxCntLable.Text = "0";
            TxCntLable.Text = "0";
        }

        private void UartConn_Click(object sender, EventArgs e)
        {
            if (serialPort1.IsOpen)
            {
                try
                {
                    serialPort1.Close();
                }
                catch 
                { }
                UartConn.Text = "连接";
                ComBox.Enabled = true;
                ComBaudBox.Enabled = true;
                ModeWriteConf.Enabled = false;
                ModeReadConf.Enabled = false;
                ModeReset.Enabled = false;
                ModeRst.Enabled = false;
                SendTxBuf.Enabled = false;
            }
            else
            {
                if (ComBox.Text.Length > 0)
                {
                    port_name = ComBox.Text;
                    serialPort1.PortName = port_name;
                    int spend = Convert.ToInt32(ComBaudBox.Text);
                    serialPort1.BaudRate = spend;
                    //serialPort1.Encoding = Encoding.GetEncoding("GB2312");
                    try
                    {
                        serialPort1.Open();
                        serialPort1.DiscardInBuffer();
                        serialPort1.DiscardOutBuffer();
                        //serialPort1.ReadBufferSize = 512;
                        UartConn.Text = "断开";
                        ComBox.Enabled = false;
                        ComBaudBox.Enabled = false;
                        ModeWriteConf.Enabled = true;
                        ModeReadConf.Enabled = true;
                        ModeReset.Enabled = true;
                        ModeRst.Enabled = true;
                        SendTxBuf.Enabled = true;
                    }
                    catch
                    {
                        UartConn.Text = "连接";
                        ComBox.Enabled = true;
                        ComBaudBox.Enabled = true;
                        ModeWriteConf.Enabled = false;
                        ModeReadConf.Enabled = false;
                        ModeReset.Enabled = false;
                        ModeRst.Enabled = false;
                        SendTxBuf.Enabled = false;
                        MessageBox.Show("串口打开失败");
                    }
                }
                else
                    MessageBox.Show("请选择一个串口");
            }
        }

        private void RxText_TextChanged(object sender, EventArgs e)
        {
            RxText.SelectionStart = RxText.TextLength;
            RxText.ScrollToCaret();
        }

        private void TxText_TextChanged(object sender, EventArgs e)
        {
            TxText.SelectionStart = TxText.TextLength;
            TxText.ScrollToCaret();
        }

        private void ModeReadConf_Click(object sender, EventArgs e)
        {
            //读配置命令
            Byte[] dat = { 0x02, 0x04 };
            UartSend(dat);
        }

        private void ModeRst_Click(object sender, EventArgs e)
        {
            //重启命令
            Byte[] dat = { 0x02, 0x08 };
            UartSend(dat);
        }

        private void ModeReset_Click(object sender, EventArgs e)
        {
            //重置命令
            Byte[] dat = { 0x02, 0x09 };
            UartSend(dat);
        }

        private void ModeWriteConf_Click(object sender, EventArgs e)
        {
            //写配置
            Byte[] dat = new byte[19];
            dat[0] = 19;
            dat[1] = 0x03;  //修改配置
            //信道0-125
            try
            {
                if (ModeRFchHex.Checked)
                {
                    if (Convert.ToInt32(ModeRFchBox.Text, 16) > 125)
                    {
                        dat[2] = 0xff;
                    }
                    else
                    {
                        dat[2] = Convert.ToByte(ModeRFchBox.Text, 16);
                    }
                }
                else
                {
                    if (Convert.ToInt32(ModeRFchBox.Text, 16) > 125)
                    {
                        dat[2] = 0xff;
                    }
                    else
                    {
                        dat[2] = Convert.ToByte(ModeRFchBox.Text);
                    }
                }
            }
            catch
            {
                dat[2] = 0xff;
            }
            //nRF_CONFIG
            try
            {
                if (ModeCRCxBox.SelectedIndex == 0)
                {
                    dat[3] = 0x0b;
                }
                else if (ModeCRCxBox.SelectedIndex == 1)
                {
                    dat[3] = 0x0f;
                }
                else
                {
                    dat[3] = 0x0f;
                }
            } catch
            {
                dat[3] = 0x0f;
            }
            //nRF_RF_SETUP
            try
            {
                if (ModeRFSpeedBox.SelectedIndex == 0)
                {
                    dat[4] = 0x07;
                }
                else if (ModeRFSpeedBox.SelectedIndex == 1)
                {
                    dat[4] = 0x0f;
                }
                else
                {
                    dat[4] = 0x0f;
                }
            }
            catch
            {

                dat[4] = 0x0f;
            }
            //发送地址
            int cnt = 0;
            try
            {
                if (ModeTxAddrBox.Text.Length == 10)
                {
                    dat[5] = 0xff;
                    for (cnt = 0; cnt < 5; cnt++)
                    {
                        dat[cnt + 6] = Convert.ToByte(ModeTxAddrBox.Text.Substring(cnt * 2, 2), 16);
                    }
                }
                else
                {
                    dat[5] = 0x00;
                }
            }
            catch
            {
                dat[5] = 0x00;
            }
            //接收地址
            try
            {
                if (ModeRxAddrBox.Text.Length == 10)
                {
                    dat[11] = 0xff;
                    for (cnt = 0; cnt < 5; cnt++)
                    {
                        dat[cnt + 12] = Convert.ToByte(ModeRxAddrBox.Text.Substring(cnt * 2, 2), 16);
                    }
                }
                else
                {
                    dat[11] = 0x00;
                }
            }
            catch
            {
                dat[11] = 0x00;
            }
            //波特率
            try
            {
                dat[17] = Convert.ToByte(ModeBaudBox.SelectedIndex);
            }
            catch
            {
                dat[17] = 0xff;
            }
            if (dat[17] > 8) dat[17] = 0xff;
            //nRF_SETUP_RETR
            try
            {
                if ((SetupRetrH.SelectedIndex >= 0) && (SetupRetrH.SelectedIndex <= 15))
                {
                    dat[18] = Convert.ToByte(0x10 * (SetupRetrH.SelectedIndex));
                    if ((SetupRetrL.SelectedIndex >= 0) && (SetupRetrL.SelectedIndex <= 15))
                    {
                        dat[18] += Convert.ToByte(SetupRetrL.SelectedIndex);
                    }
                    else
                    {
                        dat[18] = 0x1a;
                    }
                }
                else
                {
                    dat[18] = 0x1a;
                }
            }
            catch
            {
                dat[18] = 0x1a;
            }
            UartSend(dat);
        }

        private void SendTxBuf_Click(object sender, EventArgs e)
        {
            int cnt;
            if (HexModeCheck.Checked)
            {
                //16进制模式
                if (TxBufText.Text.Length % 2 != 0)
                {
                    MessageBox.Show("参数错误");
                    return;
                }
                try
                {
                    Byte[] dat = new Byte[((TxBufText.Text.Length / 2)+2)];
                    for (cnt = 0; cnt < (TxBufText.Text.Length / 2); cnt++)
                    {
                        dat[cnt + 2] = Convert.ToByte(TxBufText.Text.Substring(cnt * 2, 2), 16);
                    }
                    dat[0] = Convert.ToByte(dat.Length);
                    dat[1] = 0x01;  //发送数据
                    TxText.Text += DateTime.Now.ToLongTimeString().ToString();
                    TxText.Text += " HEX\r\n";
                    for (cnt = 0; cnt < (TxBufText.Text.Length / 2); cnt++)
                    {
                        TxText.Text += dat[cnt + 2].ToString("X2").ToUpper();
                        TxText.Text += " ";
                    }
                    TxText.Text += "\r\n";
                    UartSend(dat);
                }
                catch
                {
                    MessageBox.Show("参数错误");
                }
            }
            else
            {
                //字符串模式
                Byte[] temp = Encoding.Convert(Encoding.GetEncoding("utf-8"),Encoding.GetEncoding(TextEncodeBox.Text), Encoding.GetEncoding("utf-8").GetBytes(TxBufText.Text));
                if (temp.Length > 32)
                {
                    MessageBox.Show("数据过长\r\n一次发送最多32字节");
                    return;
                }
                Byte[] dat = new Byte[(temp.Length) + 2];
                for (cnt = 0; cnt < temp.Length; cnt++)
                {
                    dat[cnt + 2] = temp[cnt];
                }
                TxText.Text += DateTime.Now.ToLongTimeString().ToString();
                TxText.Text += " TEXT\r\n";
                TxText.Text += TxBufText.Text;
                TxText.Text += "\r\n";
                dat[0] = Convert.ToByte(dat.Length);
                dat[1] = 0x01;
                UartSend(dat);
            }
        }

        private void ModeRFchHex_CheckedChanged(object sender, EventArgs e)
        {
            if (ModeRFchHex.Checked)
            {
                ModeRFchHex.Text = "0x";
                try
                {
                    ModeRFchBox.Text = Convert.ToString(Convert.ToInt32(ModeRFchBox.Text), 16);
                }
                catch { }
            }
            else
            {
                ModeRFchHex.Text = "Hex"; try
                {
                    ModeRFchBox.Text = Convert.ToString(Convert.ToInt32(ModeRFchBox.Text, 16));
                }
                catch { }
            }
        }

        private void DataCntNRF_CheckedChanged(object sender, EventArgs e)
        {
            if (DataCntNRF.Checked)
            {
                DataCntNRF.Text = "只记录nRF数据包";
            }
            else
            {
                DataCntNRF.Text = "记录串口数据包";
            }
        }

        private void serialPort1_DataReceived(object sender, System.IO.Ports.SerialDataReceivedEventArgs e)
        {
            //读取数据
            Byte[] temp = new Byte[4096];
            
            int cnt = serialPort1.Read(temp, 0, serialPort1.BytesToRead);
            for (int i = 0; i < cnt; i++)
            {
                RecvBuf[RecvCnt] = temp[i];
                RecvCnt++;
            }
            //int cnt = serialPort1.ReadBufferSize;
            //if ((RecvCnt + cnt) > 3072)
            //{
            //    RecvCnt = 0;    //数据超过缓存则丢失之前的数据
            //}
            //RecvCnt += serialPort1.Read(RecvBuf, RecvCnt, cnt);
            //初始化回调函数
            RecvInvoke mi = new RecvInvoke(RecvData);
            //回调
            BeginInvoke(mi);
        }

        private void serialPort1_ErrorReceived(object sender, System.IO.Ports.SerialErrorReceivedEventArgs e)
        {
            RecvError me = new RecvError(RxError);
            BeginInvoke(me);
            return;
        }

        private void ModeIDBox_TextChanged(object sender, EventArgs e)
        {
            if (ModeIDBox.Text != "")
                this.Text = "nRF24L01 - " + ModeIDBox.Text;
        }

        private void EncodeLock_CheckedChanged(object sender, EventArgs e)
        {
            if (EncodeLock.Checked)
            {
                TextEncodeBox.Enabled = false;
            }
            else
            {
                TextEncodeBox.Enabled = true;
                MessageBox.Show("报文编码格式根据单片机编码格式选择");
            }
        }
    }
}
